package client

import (
	"encoding/json"
	"errors"
	"io"
	"net/http"
	"strings"
	"time"
)

// API is used to store the API-access-related data, it is generated in the NewAPIConfig function.
type API struct {
	ClientID     string
	ClientSecret string
	URL          string
	PushToken    string
	DeviceType   string
	DeviceToken  string
}

//AccessCredentials is used to handle the received Access_token JSON object after successfully login.
type AccessCredentials struct {
	AccessToken  string `json:"access_token"`
	ExpireIn     int32  `json:"expires_in"`
	Host         string `json:"host"`
	NewUser      string `json:"new_user"`
	RefreshToken string `json:"refresh_token"`
	Scope        string `json:"scope"`
	TokenType    string `json:"token_type"`
	UUID         string `json:"uuid"`
}

//ToJSONString is used to return the AccessCredentials object in a JSON string
func (accessCredentials *AccessCredentials) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(accessCredentials)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//APIResponse is used to handle the received JSON object in response
type APIResponse struct {
	Type    string          `json:"Type"`
	Content json.RawMessage `json:"Content"`
}

//Feedback is used to handle the Content part of the received JSON object which contains code and reason two fields.
type Feedback struct {
	Code   int32  `json:"code"`
	Reason string `json:"reason"`
}

//ToJSONString is used to return the Feedback object in a JSON string
func (feedback *Feedback) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(feedback)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//SessionOrder is used in the GetSessions function to request an array of session objects of a User, it contains the User
// ID and the minimum update order of sessions of that user you want to request
type SessionOrder struct {
	UserID      string `json:"UserID"`
	UpdateOrder int32  `json:"UpdateOrder"`
}

//SessionOrderArray is an array of SessionOrder objects used in the GetSessions function
type SessionOrderArray struct {
	SessionOrders []*SessionOrder `json:"SessionOrders"`
}

//InsertSessionOrder inserts a new SessionOrder object into the SessionOrders field of the SessionOrdersArray object and returns
// the array length of the SessionOrders field after the insertion.
func (sessionOrderArray *SessionOrderArray) InsertSessionOrder(newElement *SessionOrder) bool {
	// if isValid, _ := checkSessionOrderObj(newElement); !isValid {
	// 	return false
	// }
	sessionOrderArray.SessionOrders = append(sessionOrderArray.SessionOrders, newElement)
	return true
}

//GetSessionOrdersLength returns the array length of the SessionOrders fields in an SessionOrdersArray object.
func (sessionOrderArray *SessionOrderArray) GetSessionOrdersLength() int32 {
	return int32(len(sessionOrderArray.SessionOrders))
}

//GetSessionOrderAt returns one SessionOrder object in the SessionOrders field (Array) of an SessionOrderArray object by index.
func (sessionOrderArray *SessionOrderArray) GetSessionOrderAt(index int32) (*SessionOrder, error) {
	if index < 0 || index >= int32(len(sessionOrderArray.SessionOrders)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	sessionOrder := sessionOrderArray.SessionOrders[index]
	return sessionOrder, nil
}

//Data stands for Data field of the Recording struct
type Data struct {
	HeartRate    int32   `json:"heartRate,omitempty"`
	AmbientTemp  float64 `json:"ambientTemp,omitempty"`
	BatteryLevel int32   `json:"batteryLevel,omitempty"`
	SurfaceTemp  float64 `json:"surfaceTemp,omitempty"`
}

//Recording is a structure containing all the information of a recording.
type Recording struct {
	ID        string     `json:"ID"`
	SessionID string     `json:"SessionID,omitempty"`
	UserID    string     `json:"UserID,omitempty"`
	Type      int32      `json:"Type,omitempty"`
	Data      *Data      `json:"Data,omitempty"` //should be taken care of on the client side as this part changes
	DataStr   string     `json:"DataStr,omitempty"`
	FileExt   string     `json:"FileExt,omitempty"`
	Created   float64    `json:"Created,omitempty"`
	Updated   float64    `json:"Updated,omitempty"`
	State     string     `json:"State,omitempty"`
	TrackData *TrackData `json:"TrackData,omitempty"`
	URLs      []string   `json:"URLs,omitempty"`
}

//ToJSONString is used to return the Recording object in a JSON string
func (recording *Recording) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(recording)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertURL inserts a new URL string into the URLs field of the Recording object and returns
// the array length of the URLs field after the insertion.
func (recording *Recording) InsertURL(newURL string) bool {
	// if newURL == "" {
	// 	return false
	// }
	recording.URLs = append(recording.URLs, newURL)
	return true
}

//GetURLsLength returns the array length of the URLs fields in a Recording object.
func (recording *Recording) GetURLsLength() int32 {
	return int32(len(recording.URLs))
}

//GetURLAt returns one URL in the URLs field (Array) of a Recording object by index.
func (recording *Recording) GetURLAt(index int32) (string, error) {
	if index < 0 || index >= int32(len(recording.URLs)) {
		return "", feedbackStr(102, "Index beyonds the borders of the array.")
	}
	url := recording.URLs[index]
	return url, nil
}

//TrackData stand for the "TrackData" field of the Recording object.
type TrackData struct {
	SerialMeasuringDevice   string `json:"SerialMeasuringDevice,omitempty"`
	FirmwareMeasuringDevice string `json:"FirmwareMeasuringDevice,omitempty"`
}

//SessionArray stands for an array of session objects used in the parameter list and return statement to avoid the array issue.
type SessionArray struct {
	Sessions []*Session `json:"Sessions"`
}

//ToJSONString is used to return the SessionArray object in a JSON string
func (sesArray *SessionArray) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(sesArray)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertSession inserts a new Session object into the Sessions field of the SessionArray object and returns
// the array length of the Sessions field after the insertion.
func (sesArray *SessionArray) InsertSession(newSession *Session) bool {
	// if isValid, _ := checkSessionObj(newSession); !isValid {
	// 	return false
	// }
	sesArray.Sessions = append(sesArray.Sessions, newSession)
	return true
}

//GetSessionsLength returns the length of the array in the Sessions field of a SessionArray object.
func (sesArray *SessionArray) GetSessionsLength() int32 {
	return int32(len(sesArray.Sessions))
}

//GetSessionAt returns one Session object in the Sessions field (Array) of a SessionArray object by index.
func (sesArray *SessionArray) GetSessionAt(index int32) (*Session, error) {
	if index < 0 || index >= int32(len(sesArray.Sessions)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	session := sesArray.Sessions[index]
	return session, nil
}

//Session stands for a session in the App.
type Session struct {
	ID          string       `json:"ID"`
	UserID      string       `json:"UserId"`
	CaregiverID string       `json:"CaregiverID"`
	Created     float64      `json:"Created"`
	Latitude    float64      `json:"Latitude"`
	Longitude   float64      `json:"Longitud"`
	Location    string       `json:"Location"`
	Note        string       `json:"Note"`
	DeviceID    string       `json:"DeviceId,omitempty"`
	DeviceModel string       `json:"DeviceModel,omitempty"`
	AppVersion  string       `json:"AppVersion"`
	UpdateOrder int32        `json:"UpdateOrder,omitempty"`
	Shareable   bool         `json:"Shareable,omitempty"`
	Recordings  []*Recording `json:"Recordings"`
	AppType     string       `json:"AppType"`
	//Sequence   int64       `json:"Sequence"`
}

//ToJSONString is used to return the Session object in a JSON string
func (session *Session) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(session)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertRecording inserts a new Recording object into the Recordings field of the Session object and returns
// the array length of the Recordings field after the insertion.
func (session *Session) InsertRecording(newRecording *Recording) bool {
	// if newRecording == nil {
	// 	return false
	// }
	session.Recordings = append(session.Recordings, newRecording)
	return true
}

//GetRecordingsLength returns the length of the array in the Recordings field of a Session object.
func (session *Session) GetRecordingsLength() int32 {
	return int32(len(session.Recordings))
}

//GetRecordingAt returns one Recording object in the Recordings field (Array) of a Session object by index.
func (session *Session) GetRecordingAt(index int32) (*Recording, error) {
	if index < 0 || index >= int32(len(session.Recordings)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	rec := session.Recordings[index]
	return rec, nil
}

//RecordingURL stands for one recording in the CreatSessionsResponse structure.
type RecordingURL struct {
	ID   string   `json:"ID"`
	URLs []string `json:"URLs"`
}

//ToJSONString is used to return the RecordingURL object in a JSON string
func (recordingURL *RecordingURL) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(recordingURL)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertURL inserts a new URL string into the URLs field of the RecordingURL object and returns
// the array length of the URLs field after the insertion.
func (recordingURL *RecordingURL) InsertURL(newURL string) bool {
	// if newURL == "" {
	// 	return false
	// }
	recordingURL.URLs = append(recordingURL.URLs, newURL)
	return true
}

//GetURLsLength returns the length of the array in the URLs field of a RecordingURL object.
func (recordingURL *RecordingURL) GetURLsLength() int32 {
	return int32(len(recordingURL.URLs))
}

//GetURLAt returns one URL in the URLs field (Array) of a RecordingURL object by index.
func (recordingURL *RecordingURL) GetURLAt(index int32) (string, error) {
	if index < 0 || index >= int32(len(recordingURL.URLs)) {
		return "", feedbackStr(102, "Index beyonds the borders of the array.")
	}
	url := recordingURL.URLs[index]
	return url, nil
}

//UploadSession stands for the upload session part of the NewSessionsResponse
type UploadSession struct {
	SessionID   string `json:"SessionID"`
	UpdateOrder int32  `json:"UpdateOrder"`
}

//ToJSONString is used to return the UploadSession object in a JSON string
func (uploadSession *UploadSession) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(uploadSession)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//UploadSessionArray stands for a array of the UploadSession object.
type UploadSessionArray struct {
	UploadSessions []*UploadSession `json:"UploadSessions"`
}

//ToJSONString is used to return the UploadSessionArray object in a JSON string
func (uploadSessions *UploadSessionArray) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(uploadSessions)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertUploadSession inserts a new UploadSession object into the UploadSessions field of the UploadSessionArray object and returns
// the array length of the UploadSessions field after the insertion.
func (uploadSessions *UploadSessionArray) InsertUploadSession(uploadSession *UploadSession) bool {
	// if isValid, _ := checkUploadSessionObj(uploadSession); !isValid {
	// 	return false
	// }
	uploadSessions.UploadSessions = append(uploadSessions.UploadSessions, uploadSession)
	return true
}

//GetUploadSessionsLength returns the length of the array in the UploadSessions field of a UploadSessionArray object.
func (uploadSessions *UploadSessionArray) GetUploadSessionsLength() int32 {
	return int32(len(uploadSessions.UploadSessions))
}

//GetUploadSessionAt returns one UploadSession object in the UploadSessions field (Array) of a UploadSessionArray object by index.
func (uploadSessions *UploadSessionArray) GetUploadSessionAt(index int32) (*UploadSession, error) {
	if index < 0 || index >= int32(len(uploadSessions.UploadSessions)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	uploadSession := uploadSessions.UploadSessions[index]
	return uploadSession, nil
}

//NewSessionsResponse stands for the Content part of the response when create session successfully.
type NewSessionsResponse struct {
	RecordingURLs     []*RecordingURL  `json:"Recordings"`
	UploadSessions    []*UploadSession `json:"UploadSessions"`
	SessionShortLinks []string         `json:"SessionShortLinks"`
}

//ToJSONString is used to return the NewSessionsResponse object in a JSON string
func (newSessionsResponse *NewSessionsResponse) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(newSessionsResponse)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertRecordingURL inserts a new RecordingURL object into the RecordingURLs field of the NewSessionsResponse object and returns
// the array length of the RecordingURLs field after the insertion.
func (newSessionsResponse *NewSessionsResponse) InsertRecordingURL(recordingURL *RecordingURL) bool {
	// if isValid, _ := checkRecordingURLObj(recordingURL); !isValid {
	// 	return false
	// }
	newSessionsResponse.RecordingURLs = append(newSessionsResponse.RecordingURLs, recordingURL)
	return true
}

//InsertUploadSession inserts a new UploadSession object into the UploadSessions field of the NewSessionsResponse object and returns
// the array length of the UploadSessions field after the insertion.
func (newSessionsResponse *NewSessionsResponse) InsertUploadSession(uploadSession *UploadSession) bool {
	// if isValid, _ := checkUploadSessionObj(uploadSession); !isValid {
	// 	return false
	// }
	newSessionsResponse.UploadSessions = append(newSessionsResponse.UploadSessions, uploadSession)
	return true
}

//InsertSessionShortLink inserts a new SessionShortLinks string into the SessionShortLinks field of the NewSessionsResponse object and returns
// the array length of the SessionShortLinks field after the insertion.
func (newSessionsResponse *NewSessionsResponse) InsertSessionShortLink(sessionShortLink string) bool {
	// if sessionShortLink == "" {
	// 	return false
	// }
	newSessionsResponse.SessionShortLinks = append(newSessionsResponse.SessionShortLinks, sessionShortLink)
	return true
}

//GetRecordingURLsLength returns the length of the array in the RecordingURLs field of a NewSessionsResponse object.
func (newSessionsResponse *NewSessionsResponse) GetRecordingURLsLength() int32 {
	return int32(len(newSessionsResponse.RecordingURLs))
}

//GetUploadSessionsLength returns the length of the array in the UploadSessions field of a NewSessionsResponse object.
func (newSessionsResponse *NewSessionsResponse) GetUploadSessionsLength() int32 {
	return int32(len(newSessionsResponse.UploadSessions))
}

//GetSessionShortLinksLength returns the length of the array in the SessionShortLinks field of a NewSessionsResponse object.
func (newSessionsResponse *NewSessionsResponse) GetSessionShortLinksLength() int32 {
	return int32(len(newSessionsResponse.SessionShortLinks))
}

//GetRecordingURLAt returns one RecordingURL object in the RecordingURLs field (Array) of a NewSessionsResponse object by index.
func (newSessionsResponse *NewSessionsResponse) GetRecordingURLAt(index int32) (*RecordingURL, error) {
	if index < 0 || index >= int32(len(newSessionsResponse.RecordingURLs)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	recordingURL := newSessionsResponse.RecordingURLs[index]
	return recordingURL, nil
}

//GetUploadSessionAt returns one UploadSession object in the UploadSessions field (Array) of a NewSessionsResponse object by index.
func (newSessionsResponse *NewSessionsResponse) GetUploadSessionAt(index int32) (*UploadSession, error) {
	if index < 0 || index >= int32(len(newSessionsResponse.UploadSessions)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	uploadSessions := newSessionsResponse.UploadSessions[index]
	return uploadSessions, nil
}

//GetSessionShortLinkAt returns one ShortLink in the SessionShortLinks field (Array) of a NewSessionsResponse object by index.
func (newSessionsResponse *NewSessionsResponse) GetSessionShortLinkAt(index int32) (string, error) {
	if index < 0 || index >= int32(len(newSessionsResponse.SessionShortLinks)) {
		return "", feedbackStr(102, "Index beyonds the borders of the array.")
	}
	shortlink := newSessionsResponse.SessionShortLinks[index]
	return shortlink, nil
}

//ID stands for a structure which only has an ID field.
type ID struct {
	ID string `json:"ID"`
}

//IDArray stands for an array of ID object.
type IDArray struct {
	IDs []*ID `json:"IDs"`
}

//InsertID inserts a new ID object into the IDs field of the IDArray object and returns
// the array length of the IDs field after the insertion.
func (idArray *IDArray) InsertID(newID *ID) bool {
	// if isValid, _ := checkIDObj(newID); !isValid {
	// 	return false
	// }
	idArray.IDs = append(idArray.IDs, newID)
	return true
}

//GetIDsLength returns the length of the array in the IDs field of a IDArray object.
func (idArray *IDArray) GetIDsLength() int32 {
	return int32(len(idArray.IDs))
}

//GetIDAt returns one ID object in the IDs field (Array) of a IDArray object by index.
func (idArray *IDArray) GetIDAt(index int32) (*ID, error) {
	if index < 0 || index >= int32(len(idArray.IDs)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	id := idArray.IDs[index]
	return id, nil
}

//ShortLinkObjArray stands for an array of ShortLink objects.
type ShortLinkObjArray struct {
	ShortLinkObjs []*ShortLinkObj `json:"ShortLinkObjs"`
}

//ToJSONString is used to return the ShortLinkObjArray object in a JSON string
func (shortLinkObjArray *ShortLinkObjArray) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(shortLinkObjArray)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertShortLinkObj inserts a new ShortLinkObj object into the ShortLinkObjs field of the ShortLinkObjArray object and returns
// the array length of the ShortLinkObjs field after the insertion.
func (shortLinkObjArray *ShortLinkObjArray) InsertShortLinkObj(shortLinkObj *ShortLinkObj) bool {
	// if isValid, _ := checkShortLinkObj(shortLinkObj); !isValid {
	// 	return false
	// }
	shortLinkObjArray.ShortLinkObjs = append(shortLinkObjArray.ShortLinkObjs, shortLinkObj)
	return true
}

//GetShortLinkObjsLength returns the length of the array in the ShortLinkObjs field of a ShortLinkObjArray object.
func (shortLinkObjArray *ShortLinkObjArray) GetShortLinkObjsLength() int32 {
	return int32(len(shortLinkObjArray.ShortLinkObjs))
}

//GetShortLinkObjAt returns one ID object in the ShortLinkObjs field (Array) of a ShortLinkObjArray object by index.
func (shortLinkObjArray *ShortLinkObjArray) GetShortLinkObjAt(index int32) (*ShortLinkObj, error) {
	if index < 0 || index >= int32(len(shortLinkObjArray.ShortLinkObjs)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	shortLink := shortLinkObjArray.ShortLinkObjs[index]
	return shortLink, nil
}

//ShortLinkObj is used for return the shortlink for a specific SessionID from API
type ShortLinkObj struct {
	SessionID string `json:"SessionID"`
	ShortLink string `json:"ShortLink"`
	SeqNo     int32  `json:"SeqNo"`
}

//ToJSONString is used to return the ShortLinkObj object in a JSON string
func (shortLinkObj *ShortLinkObj) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(shortLinkObj)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//AuthData stands for the authData part of the SocialLogin struct
type AuthData struct {
	AuthTokenSecret string `json:"AuthTokenSecret"`
	AuthMethod      string `json:"AuthMethod"`
	AuthToken       string `json:"AuthToken"`
	ClientID        string `json:"ClientID"`
	ClientSecret    string `json:"ClientSecret"`
}

//SocialUser stands for the user part of the SocialLogin struct
type SocialUser struct {
	Mail       string     `json:"Mail"`
	FirstName  string     `json:"First_name"`
	LastName   string     `json:"Last_name"`
	Dob        float64    `json:"Dob"`
	Gender     int32      `json:"Gender"`
	Country    string     `json:"Country"`
	BioDetails *BioDetail `json:"Bio_details"`
}

//SocialAccount stands for the user from the third-party application, such as facebook
type SocialAccount struct {
	User     *SocialUser `json:"User"`
	AuthData *AuthData   `json:"authData"`
}

//BioDetail stands for the BioDetails field in the User structure
type BioDetail struct {
	UseMetric bool    `json:"UseMetric,omitempty"`
	Weight    float64 `json:"Weight,omitempty"`
	Height    float64 `json:"Height,omitempty"`
	//AuditoryID string  `json:"AuditoryID,omitempty"`
}

//UserArray stands for an array of User object.
type UserArray struct {
	Users []*User `json:"Users"`
}

//ToJSONString is used to return the UserArray object in a JSON string
func (users *UserArray) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(users)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertUser inserts a new User object into the Users field of the UserArray object and returns
// the array length of the Users field after the insertion.
func (users *UserArray) InsertUser(newUser *User) bool {
	// if isValid, _ := checkUserObj(newUser); !isValid {
	// 	return false
	// }
	users.Users = append(users.Users, newUser)
	return true
}

//GetUsersLength returns the length of the array in the Users field of a UserArray object.
func (users *UserArray) GetUsersLength() int32 {
	return int32(len(users.Users))
}

//GetUserAt returns one User object in the Users field (Array) of a UserArray object by index.
func (users *UserArray) GetUserAt(index int32) (*User, error) {
	if index < 0 || index >= int32(len(users.Users)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	user := users.Users[index]
	return user, nil
}

//User stands for the received User JSON Object
type User struct {
	ID             string        `json:"Id"`
	Mail           string        `json:"Mail"`
	FirstName      string        `json:"First_name"`
	LastName       string        `json:"Last_name"`
	Dob            float64       `json:"Dob"`
	Gender         int32         `json:"Gender"`
	Country        string        `json:"Country"`
	Password       string        `json:"Password,omitempty"`
	State          string        `json:"State"`
	CarePermission string        `json:"CarePermission,omitempty"`
	NotifyRecord   bool          `json:"NotifyRecord"`
	BioDetails     *BioDetail    `json:"Bio_details,omitempty"`
	PriDependents  []*User       `json:"PriDependents,omitempty"`
	SecDependents  []*PrimaryDep `json:"SecDependents,omitempty"`
	PendDependents []*User       `json:"PendDependents,omitempty"`
	Caregivers     []*User       `json:"Caregivers,omitempty"`
	//AuditoryID     string        `json:"AuditoryID,omitempty"`
}

//ToJSONString is used to return the User object in a JSON string
func (user *User) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(user)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertPriDependent inserts a new User object into the PriDependents field of the User object and returns
// the array length of the PriDependents field after the insertion.
func (user *User) InsertPriDependent(newUser *User) bool {
	// if isValid, _ := checkUserObj(newUser); !isValid {
	// 	return false
	// }
	user.PriDependents = append(user.PriDependents, newUser)
	return true
}

//InsertSecDependent inserts a new PrimaryDep object into the SecDependents field of the User object and returns
// the array length of the SecDependents field after the insertion.
func (user *User) InsertSecDependent(primaryDep *PrimaryDep) bool {
	// if isValid, _ := checkPrimaryDepObj(primaryDep); !isValid {
	// 	return false
	// }
	user.SecDependents = append(user.SecDependents, primaryDep)
	return true
}

//InsertPendDependent inserts a new User object into the PendDependents field of the User object and returns
// the array length of the PendDependents field after the insertion.
func (user *User) InsertPendDependent(newUser *User) bool {
	// if isValid, _ := checkUserObj(newUser); !isValid {
	// 	return false
	// }
	user.PendDependents = append(user.PendDependents, newUser)
	return true
}

//InsertCaregiver inserts a new User object into the Caregivers field of the User object and returns
// the array length of the Caregivers field after the insertion.
func (user *User) InsertCaregiver(newUser *User) bool {
	// if isValid, _ := checkUserObj(newUser); !isValid {
	// 	return false
	// }
	user.Caregivers = append(user.Caregivers, newUser)
	return true
}

//GetPriDependentsLength returns the length of the array in the PriDependents field of a User object.
func (user *User) GetPriDependentsLength() int32 {
	return int32(len(user.PriDependents))
}

//GetSecDependentsLength returns the length of the array in the SecDependents field of a User object.
func (user *User) GetSecDependentsLength() int32 {
	return int32(len(user.SecDependents))
}

//GetPendDependentsLength returns the length of the array in the PendDependents field of a User object.
func (user *User) GetPendDependentsLength() int32 {
	return int32(len(user.PendDependents))
}

//GetCaregiversLength returns the length of the array in the Caregivers field of a User object.
func (user *User) GetCaregiversLength() int32 {
	return int32(len(user.Caregivers))
}

//GetPriDependentAt returns one User object in the PriDependents field (Array) of a User object by index.
func (user *User) GetPriDependentAt(index int32) (*User, error) {
	if index < 0 || index >= int32(len(user.PriDependents)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	aUser := user.PriDependents[index]
	return aUser, nil
}

//GetSecDependentAt returns one PrimaryDep object in the SecDependents field (Array) of a User object by index.
func (user *User) GetSecDependentAt(index int32) (*PrimaryDep, error) {
	if index < 0 || index >= int32(len(user.SecDependents)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	aUser := user.SecDependents[index]
	return aUser, nil
}

//GetPendDependentAt returns one User object in the PendDependents field (Array) of a User object by index.
func (user *User) GetPendDependentAt(index int32) (*User, error) {
	if index < 0 || index >= int32(len(user.PendDependents)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	aUser := user.PendDependents[index]
	return aUser, nil
}

//GetCaregiverAt returns one User object in the Caregivers field (Array) of a User object by index.
func (user *User) GetCaregiverAt(index int32) (*User, error) {
	if index < 0 || index >= int32(len(user.Caregivers)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	aUser := user.Caregivers[index]
	return aUser, nil
}

//PrimaryDep stands for the primary dependents
type PrimaryDep struct {
	Primary    *User   `json:"Primary"`
	Dependents []*User `json:"Dependents"`
}

//InsertDependent inserts a new User object into the Dependents field of the PrimaryDep object and returns
// the array length of the Dependents field after the insertion.
func (priDep *PrimaryDep) InsertDependent(newUser *User) bool {
	// if isValid, _ := checkUserObj(newUser); !isValid {
	// 	return false
	// }
	priDep.Dependents = append(priDep.Dependents, newUser)
	return true
}

//GetDependentsLength returns the length of the array in the Dependents field of a PrimaryDep object.
func (priDep *PrimaryDep) GetDependentsLength() int32 {
	return int32(len(priDep.Dependents))
}

//GetDependentAt returns one User object in the Dependents field (Array) of a PrimaryDep object by index.
func (priDep *PrimaryDep) GetDependentAt(index int32) (*User, error) {
	if index < 0 || index >= int32(len(priDep.Dependents)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	dependent := priDep.Dependents[index]
	return dependent, nil
}

// DependentDetail stores dependents information
type DependentDetail struct {
	UserID       string `json:"User_id"`
	Permission   string `json:"Permission,omitempty"`
	NotifyRecord bool   `json:"Notify_record,omitempty"`
	FirstName    string `json:"First_name,omitempty"`
	LastName     string `json:"Last_name,omitempty"`
}

//ToJSONString is used to return the DependentDetail object in a JSON string
func (dependentDetail *DependentDetail) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(dependentDetail)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//DependentDetailArray stands for an array of DependentDetail
type DependentDetailArray struct {
	DependentDetails []*DependentDetail `json:"DependentDetails"`
}

//ToJSONString is used to return the DependentDetailArray object in a JSON string
func (depDetailArray *DependentDetailArray) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(depDetailArray)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertDependentDetail inserts a new DependentDetail object into the DependentDetails field of the DependentDetailArray object and returns
// the array length of the DependentDetails field after the insertion.
func (depDetailArray *DependentDetailArray) InsertDependentDetail(depDetail *DependentDetail) bool {
	// if isValid, _ := checkDependentDetailObj(depDetail); !isValid {
	// 	return false
	// }
	depDetailArray.DependentDetails = append(depDetailArray.DependentDetails, depDetail)
	return true
}

//GetDependentDetailsLength returns the length of the array in the DependentDetails field of a DependentDetailArray object.
func (depDetailArray *DependentDetailArray) GetDependentDetailsLength() int32 {
	return int32(len(depDetailArray.DependentDetails))
}

//GetDependentDetailAt returns one DependentDetail object in the DependentDetails field (Array) of a DependentDetailArray object by index.
func (depDetailArray *DependentDetailArray) GetDependentDetailAt(index int32) (*DependentDetail, error) {
	if index < 0 || index >= int32(len(depDetailArray.DependentDetails)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	dependentDetail := depDetailArray.DependentDetails[index]
	return dependentDetail, nil
}

//HelpFileArray stands for an array of HelpFile
type HelpFileArray struct {
	HelpFiles []*HelpFile `json:"HelpFiles"`
}

//ToJSONString is used to return the HelpFileArray object in a JSON string
func (helpFileArray *HelpFileArray) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(helpFileArray)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//InsertHelpFile inserts a new HelpFile object into the HelpFiles field of the HelpFileArray object and returns
// the array length of the HelpFiles field after the insertion.
func (helpFileArray *HelpFileArray) InsertHelpFile(newHelpFile *HelpFile) bool {
	// if isValid, _ := checkHelpFileObj(newHelpFile); !isValid {
	// 	return false
	// }
	helpFileArray.HelpFiles = append(helpFileArray.HelpFiles, newHelpFile)
	return true
}

//GetHelpFilesLength returns the length of the array in the HelpFiles field of a HelpFileArray object.
func (helpFileArray *HelpFileArray) GetHelpFilesLength() int32 {
	return int32(len(helpFileArray.HelpFiles))
}

//GetHelpFileAt returns one HelpFile object in the HelpFiles field (Array) of a HelpFileArray object by index.
func (helpFileArray *HelpFileArray) GetHelpFileAt(index int32) (*HelpFile, error) {
	if index < 0 || index >= int32(len(helpFileArray.HelpFiles)) {
		return nil, feedbackStr(102, "Index beyonds the borders of the array.")
	}
	helpFile := helpFileArray.HelpFiles[index]
	return helpFile, nil
}

//HelpFile stores help file infomation
type HelpFile struct {
	Name       string `json:"Name"`
	Path       string `json:"Path"`
	Version    int32  `json:"Version"`
	AppVersion string `json:"AppVersion,omitempty"`
}

//ToJSONString is used to return the HelpFile object in a JSON string
func (helpFile *HelpFile) ToJSONString() (string, error) {
	jsonBytes, jsonErr := json.Marshal(helpFile)
	if jsonErr != nil {
		return "", feedbackStr(101, jsonErr.Error())
	}
	return string(jsonBytes), nil
}

//StringArray is used for using string array in return or parameter list
type StringArray struct {
	Strings []string `json:"Strings"`
}

//InsertString inserts a new string into the Strings field of the StringArray object and returns
// the array length of the Strings field after the insertion.
func (strs *StringArray) InsertString(newStr string) bool {
	// if newStr == "" {
	// 	return false
	// }
	strs.Strings = append(strs.Strings, newStr)
	return true
}

//GetStringsLength returns the length of the array in the Strings field of a StringArray object.
func (strs *StringArray) GetStringsLength() int32 {
	return int32(len(strs.Strings))
}

//GetStringAt returns one string in the Strings field (Array) of a StringArray object by index.
func (strs *StringArray) GetStringAt(index int32) (string, error) {
	if index < 0 || index >= int32(len(strs.Strings)) {
		return "", feedbackStr(102, "Index beyonds the borders of the array.")
	}
	str := strs.Strings[index]
	return str, nil
}

//apiRequestHeader is a private(internal) function which is used for generating a request header object for a specific kind of request.
func apiRequestHeader(isLogin, headerType bool, accessToken string, api *API) (*http.Header, error) {

	//check each parameter whether it is empty
	switch {
	case accessToken == "":
		return nil, feedbackStr(102, "The parameter accessToken cannot be empty.")
	case api == nil:
		return nil, feedbackStr(102, "The parameter api cannot be empty.")
	}

	//create and set the key and value pairs
	headers := http.Header{}
	headers.Set("Content-Type", "application/json; charset=utf-8")
	if isLogin {
		headers.Set("Authorization", " "+accessToken)
	} else {
		headers.Set("Authorization", " "+accessToken)
	}

	headers.Set("DeviceToken", api.DeviceToken)
	headers.Set("DeviceType", api.DeviceType)
	headers.Set("PushToken", api.PushToken)

	if headerType {
		headers.Set("ClientID", api.ClientID)
		headers.Set("ClientSecret", api.ClientSecret)
	}

	return &headers, nil
}

//apiRequest is a private(internal) function which is used for sending the request to the APIs and
// returning the response data in JSON format, the parameter category is used to distinguish the URLs
// having the same type of response.
func apiRequest(reqMethod, url string, reqBody io.Reader, headers *http.Header, category int32) (interface{}, error) {

	method := strings.ToLower(reqMethod)

	//check each parameter whether it is empty
	switch {
	case reqMethod == "":
		return nil, feedbackStr(102, "The parameter reqMethod cannot be empty.")
	case url == "":
		return nil, feedbackStr(102, "The parameter url cannot be empty.")
	case method != "get" && method != "delete" && reqBody == nil:
		return nil, feedbackStr(102, "The parameter reqBody cannot be empty except using GET or DELETE method.")
	case headers == nil:
		return nil, feedbackStr(102, "The parameter headers cannot be empty.")
	case category < 0:
		return nil, feedbackStr(102, "The parameter category cannot be negative.")
	}

	//create request based on method, url and request body
	req, reqErr := http.NewRequest(reqMethod, url, reqBody)
	if reqErr != nil {
		// handle request error
		return nil, feedbackStr(103, reqErr.Error())
	}

	//add headers to the request
	req.Header = *headers
	// tr := &http.Transport{
	// 	TLSClientConfig: &tls.Config{InsecureSkipVerify: true},
	// }

	//set client timeout
	client := &http.Client{
		Timeout: 10 * time.Second,
		// Transport: tr,
	}

	//send request to the API and get the response
	res, resErr := client.Do(req)
	// fmt.Println(res.StatusCode)
	// fmt.Println(res.Body)
	if resErr != nil {
		return nil, feedbackStr(103, resErr.Error())
	} else if res.StatusCode != 200 && res.ContentLength < 20 {
		//fmt.Println(res.StatusCode)
		errObj := Feedback{
			Code:   18,
			Reason: "Email is not valid.",
		}
		errObjStr, errObjStrErr := errObj.ToJSONString()
		if errObjStrErr != nil {
			return nil, feedbackStr(101, errObjStrErr.Error())
		}
		return nil, errors.New(errObjStr)
		//return nil, feedbackStr(103, "Some communication error happened.")
	}

	defer res.Body.Close()

	//handle the response
	resObj := APIResponse{}
	decodeErr := json.NewDecoder(res.Body).Decode(&resObj)
	if decodeErr != nil {
		return nil, feedbackStr(101, decodeErr.Error())
	}
	return convertContent(&resObj, category)

}

//convertContent is a private(internal) function which is responsible for converting the Content part of
// a response body into a corresponding struct type based on the category and the Type part of its body.
func convertContent(responseObj *APIResponse, category int32) (interface{}, error) {
	//check the parameter
	if responseObj == nil {
		return nil, feedbackStr(102, "The parameter responseObj cannot be empty.")
	}

	//Based on the value of the Type part of the JSON Object in the response,
	//convert the Content part in that JSON Object into a corresponding struct type
	switch responseObj.Type {
	case "Access_token":
		accessToken := AccessCredentials{}
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &accessToken)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		return accessToken, nil
	case "user_info":
		switch category {
		case 0:
			userInfo := User{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &userInfo)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			return userInfo, nil
		case 1:
			usersInfo := []*User{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &usersInfo)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			userArray := UserArray{
				Users: usersInfo,
			}
			return userArray, nil
		case 2:
			dependentDetails := []*DependentDetail{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &dependentDetails)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			dependentDetailArray := DependentDetailArray{
				DependentDetails: dependentDetails,
			}
			return dependentDetailArray, nil
		default:
			return nil, feedbackStr(101, "No such category number is predefined.")
		}
	case "success":
		checkResult := Feedback{}
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &checkResult)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		return checkResult, nil
	case "blob_url":
		blobURL := ""
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &blobURL)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		return blobURL, nil
	case "user_id":
		userID := ""
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &userID)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		return userID, nil
	case "recordings-url":
		switch category {
		case 0:
			feedback := NewSessionsResponse{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &feedback)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			return feedback, nil
		case 1:
			ses := []string{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &ses)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			strArray := StringArray{
				Strings: ses,
			}
			return strArray, nil
		default:
			return nil, feedbackStr(101, "No such category number is predefined.")
		}
	case "recordings":
		uploadSessions := []*UploadSession{}
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &uploadSessions)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		uploadSesArray := UploadSessionArray{
			UploadSessions: uploadSessions,
		}
		return uploadSesArray, nil
	case "sessions":
		switch category {
		case 0:
			ses := []*Session{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &ses)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			sessionArray := SessionArray{
				Sessions: ses,
			}
			return sessionArray, nil
		case 1:
			ses := Feedback{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &ses)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			return ses, nil
		case 2:
			ses := Session{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &ses)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			return ses, nil
		default:
			return nil, feedbackStr(101, "No such category number is predefined.")
		}
	case "help_files":
		switch category {
		case 0:
			files := []*HelpFile{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &files)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			helpFiles := HelpFileArray{
				HelpFiles: files,
			}
			return helpFiles, nil
		case 1:
			feedback := Feedback{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &feedback)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			return feedback, nil
		case 2:
			file := HelpFile{}
			unmarErr := json.Unmarshal([]byte(responseObj.Content), &file)
			if unmarErr != nil {
				return nil, feedbackStr(101, unmarErr.Error())
			}
			return file, nil
		default:
			return nil, feedbackStr(101, "No such category number is predefined.")
		}
	case "shortlink":
		ShortLinkList := []*ShortLinkObj{}
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &ShortLinkList)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		shortLinkObjArray := ShortLinkObjArray{
			ShortLinkObjs: ShortLinkList,
		}
		return shortLinkObjArray, nil
	case "mail_sent":
		userID := ""
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &userID)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		return userID, nil
	case "error", "error_user":
		apiErrObj := Feedback{}
		unmarErr := json.Unmarshal([]byte(responseObj.Content), &apiErrObj)
		if unmarErr != nil {
			return nil, feedbackStr(101, unmarErr.Error())
		}
		errBytes, marErr := json.Marshal(apiErrObj)
		if marErr != nil {
			return nil, feedbackStr(101, marErr.Error())
		}
		return nil, errors.New(string(errBytes))
	default:
		return nil, feedbackStr(101, "No such response type is predefined.")
	}
}

//feedbackStr is used to convert the client side error and the communicaiton error into the code/reason form, which
// makes all the error messages in the uniform format.
func feedbackStr(errCode int32, errMsg string) error {
	switch {
	case errCode == 101:
		return errors.New("{\"code\": 101, \"reason\": \"Internal error: " + errMsg + "\"}")
	case errCode == 102:
		return errors.New("{\"code\": 102, \"reason\": \"Parameter error: " + errMsg + "\"}")
	case errCode == 103:
		return errors.New("{\"code\": 103, \"reason\": \"Communication error: " + errMsg + "\"}")
	default:
		return errors.New("{\"code\": 100, \"reason\": \"Unknown error: " + errMsg + "\"}")
	}
}
